from contextlib import suppress
from RobotRaconteur.RobotRaconteurPythonUtil import NamedArrayToArray
import numpy as np
import abb_motion_program_exec as abb_exec
import RobotRaconteur as RR
import general_robotics_toolbox as rox

def rr_pose_to_abb(rr_pose):
    a = NamedArrayToArray(rr_pose)
    return abb_exec.pose(a[0][4:7]*1000.0,a[0][0:4])

def rr_inertia_to_loaddata_abb(rr_inertia, rr_inertia_pose = None):
    ii = NamedArrayToArray(rr_inertia)[0]
    # ii[0] = m
    # ii[1:4] = com
    # ii[4] = Ixx
    # ii[5] = Ixy
    # ii[6] = Ixz
    # ii[7] = Iyy
    # ii[8] = Iyz
    # ii[9] = Izz
    mass = ii[0]
    com = ii[1:4]
    I = np.array([[ii[4],ii[5],ii[6]],[ii[5],ii[7],ii[8]],[ii[6],ii[8],ii[9]]])
    w_eig,v_eig = np.linalg.eig(I)
    R = np.column_stack([v_eig[:,0], v_eig[:,1], np.cross(v_eig[:,0], v_eig[:,1])])
    aom = rox.R2q(R)
    ix, iy, iz = w_eig

    if rr_inertia_pose is not None:
        com = np.add(com, NamedArrayToArray(rr_inertia_pose[0]["position"]))
        aom = rox.quatproduct(NamedArrayToArray(rr_inertia_pose[0]["orientation"])) @ aom
    
    return abb_exec.loaddata(mass, com, aom, ix, iy, iz)

def rr_tool_to_abb(rr_tool_info):
    tcp = rr_pose_to_abb(rr_tool_info.tcp)
    ld = rr_inertia_to_loaddata_abb(rr_tool_info.inertia)
    return abb_exec.tooldata(True,tcp,ld)

def rr_payload_to_abb(rr_payload_info, rr_payload_pose):
    return rr_inertia_to_loaddata_abb(rr_payload_info.inertia, rr_payload_pose)

def rr_zone_to_abb(rr_fine_point,rr_blend_radius):
    r = rr_blend_radius * 1000.0
    #TODO use "extended" fields for external axes
    return abb_exec.zonedata(rr_fine_point,r,r,r,r,r,r)

def rr_speed_to_abb(rr_velocity):
    #TODO use "extended" for angular velocity and external axes?
    return abb_exec.speeddata(rr_velocity*1000.0, rr_velocity*10000.0, 1000, 1000)

def rr_joints_to_abb(rr_joints, rr_joint_units):
    #TODO: joint units
    #TODO: use "extended" for external axes
    return abb_exec.jointtarget(np.rad2deg(rr_joints), [6e5]*6)

def rr_robot_pose_to_abb(rr_robot_pose, cfx_robot):
    #TODO: joint units
    #TODO: use "extended" for external axes
    p = rr_pose_to_abb(rr_robot_pose.tcp_pose)
    wrist_vs_axis1 = rox.fwdkin(cfx_robot, [0,rr_robot_pose.joint_position_seed[1],
        rr_robot_pose.joint_position_seed[2],0,0,0]).p[0] < 0
    wrist_vs_lower_arm = rox.fwdkin(cfx_robot, [0,0,rr_robot_pose.joint_position_seed[2],0,0,0]).p[0] < 0
    axis_5_sign = rr_robot_pose.joint_position_seed[4] < 0
    cfx = np.packbits([axis_5_sign, wrist_vs_lower_arm, wrist_vs_axis1], bitorder='little').item()
    cd = abb_exec.confdata(
        np.floor(rr_robot_pose.joint_position_seed[0]/(np.pi/2)),
        np.floor(rr_robot_pose.joint_position_seed[3]/(np.pi/2)),
        np.floor(rr_robot_pose.joint_position_seed[5]/(np.pi/2)),
        cfx
    )
    return abb_exec.robtarget(p.trans,p.rot,cd, [6e5]*6)

cmd_get_arg_sentinel = object()
cmd_arg_no_default = object()

def cmd_get_arg(cmd, arg_name, default_value = cmd_arg_no_default):
    if isinstance(cmd, RR.VarValue):
        cmd = cmd.data

    val = getattr(cmd, arg_name, cmd_get_arg_sentinel)
    if val is cmd_get_arg_sentinel:
        freeform_args = getattr(cmd, "command_args", cmd_get_arg_sentinel)
        assert freeform_args is not cmd_get_arg_sentinel, f"Invalid command type, missing argument {arg_name}"

        val = freeform_args.get(arg_name, cmd_get_arg_sentinel)
        if val is cmd_get_arg_sentinel and default_value is not cmd_arg_no_default:
            return default_value
        assert val is not cmd_get_arg_sentinel, f"Invalid command type, missing argument {arg_name}"

    if isinstance(val, RR.VarValue):
        val = val.data

    return val

class MoveAbsJCommandConv:
    rr_types = ["experimental.robotics.motion_program.MoveAbsJCommand"]
    freeform_names = ["MoveAbsJ", "MoveAbsJCommand", "experimental.robotics.motion_program.MoveAbsJCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd, "tcp_velocity"))
        jt = rr_joints_to_abb(cmd_get_arg(cmd,"joint_position"), cmd_get_arg(cmd,"joint_units", []))
        mp.MoveAbsJ(jt, sd, zd)

class MoveJCommandConv:
    rr_types = ["experimental.robotics.motion_program.MoveJCommand"]
    freeform_names = ["MoveJ","MoveJCommand","experimental.robotics.motion_program.MoveJCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot, **kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd,"tcp_velocity"))
        rt = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_pose"), cfx_robot)
        mp.MoveJ(rt, sd, zd)

class MoveLCommandConv:
    rr_types = ["experimental.robotics.motion_program.MoveLCommand"]
    freeform_names = ["MoveL","MoveLCommand","experimental.robotics.motion_program.MoveLCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot, **kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd,"tcp_velocity"))
        rt = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_pose"), cfx_robot)
        mp.MoveL(rt, sd, zd)

class MoveCCommandConv:
    rr_types = ["experimental.robotics.motion_program.MoveCCommand"]
    freeform_names = ["MoveC","MoveCCommand","experimental.robotics.motion_program.MoveCCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot, **kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd,"tcp_velocity"))
        rt = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_pose"), cfx_robot)
        rt2 = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_via_pose"), cfx_robot)
        mp.MoveC(rt2, rt,  sd, zd)

class WaitTimeCommandConv:
    rr_types = ["experimental.robotics.motion_program.WaitTimeCommand"]
    freeform_names = ["WaitTime", "SetToolCommand", "experimental.robotics.motion_program.WaitTimeCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        mp.WaitTime(cmd_get_arg(cmd, "time"))

class SetToolCommandConv:
    rr_types = ["experimental.robotics.motion_program.SetToolCommand"]
    freeform_names = ["SetTool", "SetToolCommand", "experimental.robotics.motion_program.SetToolCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        abb_tool = rr_tool_to_abb(cmd_get_arg(cmd,"tool_info"))
        setup_args["tool"] = abb_tool

class SetPayloadCommandConv:
    rr_types = ["experimental.robotics.motion_program.SetPayloadCommand"]
    freeform_names = ["SetPayload", "SetPayloadCommand", "experimental.robotics.motion_program.SetPayloadCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        abb_payload = rr_payload_to_abb(cmd_get_arg(cmd,"payload_info"),cmd_get_arg(cmd,"payload_pose"))
        setup_args["gripload"] = abb_payload

# ABB Commands

class CirPathModeCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.CirPathModeCommand"]
    freeform_names = ["CirPathMode", "CirPathModeCommand", "experimental.abb_robot.motion_program.CirPathModeCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        rr_switch = cmd_get_arg(cmd, "switch")
        if rr_switch == 1 or rr_switch == "PathFrame":
            switch = abb_exec.CirPathModeSwitch.PathFrame
        elif rr_switch == 2 or rr_switch == "ObjectFrame":
            switch = abb_exec.CirPathModeSwitch.ObjectFrame
        elif rr_switch == 3 or rr_switch == "CirPointOri":
            switch = abb_exec.CirPathModeSwitch.CirPointOri
        elif rr_switch == 4 or rr_switch == "Wrist45":
            switch = abb_exec.CirPathModeSwitch.Wrist45
        elif rr_switch == 5 or rr_switch == "Wrist46":
            switch = abb_exec.CirPathModeSwitch.Wrist46
        elif rr_switch == 6 or rr_switch == "Wrist56":
            switch = abb_exec.CirPathModeSwitch.Wrist56
        else:
            assert False, f"Invalid CirPathModeSwitch value: {rr_switch}"
        mp.CirPathMode(switch)

class SyncMoveOnCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.SyncMoveOnCommand"]
    freeform_names = ["SyncMoveOn", "SyncMoveOnCommand", "experimental.abb_robot.motion_program.SyncMoveOnCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        mp.SyncMoveOn()

class SyncMoveOffCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.SyncMoveOffCommand"]
    freeform_names = ["SyncMoveOff", "SyncMoveOffCommand", "experimental.abb_robot.motion_program.SyncMoveOffCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        mp.SyncMoveOff()

# ABB EGM Commands

class EGMStreamConfigCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMStreamConfigCommand"]
    freeform_names = ["EGMStreamConfig", "EGMStreamConfigCommand", "experimental.abb_robot.motion_program.EGMStreamConfigCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        setup_args["egm_config"] = abb_exec.EGMStreamConfig()

def rr_egmframetype_to_abb(rr_egmframetype):
    if rr_egmframetype == 0 or rr_egmframetype == "EGM_FRAME_BASE":
        return abb_exec.egmframetype.EGM_FRAME_BASE
    elif rr_egmframetype == 1 or rr_egmframetype == "EGM_FRAME_TOOL":
        return abb_exec.egmframetype.EGM_FRAME_TOOL
    elif rr_egmframetype == 2 or rr_egmframetype == "EGM_FRAME_WOBJ":
        return abb_exec.egmframetype.EGM_FRAME_WOBJ
    elif rr_egmframetype == 3 or rr_egmframetype == "EGM_FRAME_WORLD":
        return abb_exec.egmframetype.EGM_FRAME_WORLD
    elif rr_egmframetype == 4 or rr_egmframetype == "EGM_FRAME_JOINT":
        return abb_exec.egmframetype.EGM_FRAME_JOINT
    else:
        assert False, f"Invalid egmframetype value: {rr_egmframetype}"

def rr_egmminmax_to_abb(rr_egmminmax):
    if not isinstance(rr_egmminmax, RR.VarValue):
        min_ = rr_egmminmax.min
        max_ = rr_egmminmax.max
        return abb_exec.egm_minmax(min_, max_)
    else:
        min_ = rr_egmminmax.data["min"].data
        max_ = rr_egmminmax.data["max"].data
        return abb_exec.egm_minmax(min_, max_)

class EGMJointTargetConfigCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMJointTargetConfigCommand"]
    freeform_names = ["EGMJointTargetConfig", "EGMJointTargetConfigCommand", "experimental.abb_robot.motion_program.EGMJointTargetConfigCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        J1 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J1"))
        J2 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J2"))
        J3 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J3"))
        J4 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J4"))
        J5 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J5"))
        J6 = rr_egmminmax_to_abb(cmd_get_arg(cmd, "J6"))
        max_position_deviation = cmd_get_arg(cmd, "max_position_deviation")
        max_speed_deviation = cmd_get_arg(cmd, "max_speed_deviation")
        setup_args["egm_config"] = abb_exec.EGMJointTargetConfig(J1, J2, J3, J4, J5, J6, max_position_deviation,
            max_speed_deviation)

class EGMPoseTargetConfigCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMPoseTargetConfigCommand"]
    freeform_names = ["EGMPoseTargetConfig", "EGMPoseTargetConfigCommand", "experimental.abb_robot.motion_program.EGMPoseTargetConfigCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        corr_frame = rr_pose_to_abb(cmd_get_arg(cmd, "corr_frame"))
        corr_fr_type = rr_egmframetype_to_abb(cmd_get_arg(cmd, "corr_fr_type"))
        sensor_frame = rr_pose_to_abb(cmd_get_arg(cmd, "sensor_frame"))
        sensor_fr_type = rr_egmframetype_to_abb(cmd_get_arg(cmd, "sensor_fr_type"))
        x = rr_egmminmax_to_abb(cmd_get_arg(cmd, "x"))
        y = rr_egmminmax_to_abb(cmd_get_arg(cmd, "y"))
        z = rr_egmminmax_to_abb(cmd_get_arg(cmd, "z"))
        rx = rr_egmminmax_to_abb(cmd_get_arg(cmd, "rx"))
        ry = rr_egmminmax_to_abb(cmd_get_arg(cmd, "ry"))
        rz = rr_egmminmax_to_abb(cmd_get_arg(cmd, "rz"))
        max_position_deviation = cmd_get_arg(cmd, "max_position_deviation")
        max_speed_deviation = cmd_get_arg(cmd, "max_speed_deviation")
        setup_args["egm_config"] = abb_exec.EGMPoseTargetConfig(corr_frame, corr_fr_type, sensor_frame, sensor_fr_type,
            x, y, z, rx, ry, rz, max_position_deviation, max_speed_deviation)

class EGMPathCorrectionConfigCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMPathCorrectionConfigCommand"]
    freeform_names = ["EGMPathCorrectionConfig", "EGMPathCorrectionConfigCommand", "experimental.abb_robot.motion_program.EGMPathCorrectionConfigCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        assert False, "Unsupported in motion command section"

    def add_setup_args(self, cmd, setup_args):
        sensor_frame = rr_pose_to_abb(cmd_get_arg(cmd, "sensor_frame"))
        setup_args["egm_config"] = abb_exec.EGMPathCorrectionConfig(sensor_frame)

class EGMRunJointCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMRunJointCommand"]
    freeform_names = ["EGMRunJoint", "EGMRunJointCommand", "experimental.abb_robot.motion_program.EGMRunJointCommand"]

    def apply_rr_command(self, cmd, mp, **kwargs):
        cond_time = cmd_get_arg(cmd, "cond_time")
        ramp_in_time = cmd_get_arg(cmd, "ramp_in_time")
        ramp_out_time = cmd_get_arg(cmd, "ramp_out_time")
        mp.EGMRunJoint(cond_time, ramp_in_time, ramp_out_time)

class EGMRunPoseCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMRunPoseCommand"]
    freeform_names = ["EGMRunPose", "EGMRunPoseCommand", "experimental.abb_robot.motion_program.EGMRunPoseCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot, **kwargs):
        cond_time = cmd_get_arg(cmd, "cond_time")
        ramp_in_time = cmd_get_arg(cmd, "ramp_in_time")
        ramp_out_time = cmd_get_arg(cmd, "ramp_out_time")
        offset = rr_pose_to_abb(cmd_get_arg(cmd, "offset"))
        mp.EGMRunPose(cond_time, ramp_in_time, ramp_out_time, offset)

class EGMMoveLCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMMoveLCommand"]
    freeform_names = ["EGMMoveL","EGMMoveLCommand","experimental.abb_robot.motion_program.EGMMoveLCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot, **kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd,"tcp_velocity"))
        rt = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_pose"),cfx_robot)
        mp.EGMMoveL(rt, sd, zd)

class EGMMoveCCommandConv:
    rr_types = ["experimental.abb_robot.motion_program.EGMMoveCCommand"]
    freeform_names = ["EGMMoveC","EGMMoveCCommand","experimental.abb_robot.motion_program.EGMMoveCCommand"]

    def apply_rr_command(self, cmd, mp, cfx_robot,**kwargs):
        zd = rr_zone_to_abb(cmd_get_arg(cmd,"fine_point"),cmd_get_arg(cmd,"blend_radius"))
        sd = rr_speed_to_abb(cmd_get_arg(cmd,"tcp_velocity"))
        rt = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_pose"),cfx_robot)
        rt2 = rr_robot_pose_to_abb(cmd_get_arg(cmd,"tcp_via_pose"),cfx_robot)
        mp.EGMMoveC(rt2, rt,  sd, zd)

_command_convs = dict()
_freeform_command_convs = dict()

conv_types = [
    MoveAbsJCommandConv,
    MoveJCommandConv,
    MoveLCommandConv,
    MoveCCommandConv,
    WaitTimeCommandConv,
    SetToolCommandConv,
    SetPayloadCommandConv,
    CirPathModeCommandConv,
    SyncMoveOnCommandConv,
    SyncMoveOffCommandConv,
    EGMStreamConfigCommandConv,
    EGMJointTargetConfigCommandConv,    
    EGMPoseTargetConfigCommandConv,
    EGMPathCorrectionConfigCommandConv,
    EGMRunJointCommandConv,
    EGMRunPoseCommandConv,
    EGMMoveLCommandConv,
    EGMMoveCCommandConv
]

def _init_convs():
    for c in conv_types:
        c_inst = c()
        for x in c_inst.rr_types:
            _command_convs[x] = c_inst
        for y in c_inst.freeform_names:
            _freeform_command_convs[y] = c_inst

_init_convs()

class OptionalCommandException(Exception):
    def __init__(self, message):
        super().__init__(message=message)

def get_command_conv(cmd):
    if cmd.datatype == "experimental.robotics.motion_program.FreeformCommand":
        conv = _freeform_command_convs.get(cmd.data.command_name, None)
        if conv is None:
            if cmd.data.optional:
                raise OptionalCommandException(f"Optional command {cmd.data.command_name}")
            else:
                assert False, f"Unknown command {cmd.data.command_name}"
        return conv
    else:
        conv = _command_convs.get(cmd.datatype, None)
        assert conv is not None, f"Unknown command {cmd.datatype}"
        return conv

def apply_rr_motion_command_to_mp(cmd, mp, **kwargs):
    conv = get_command_conv(cmd)
    conv.apply_rr_command(cmd, mp, **kwargs)


def rr_motion_program_to_abb(rr_mp, robot, rr_current_tool, rr_current_payload, rr_current_payload_pose):

    # Robot structure for computing confdata.cfx from joint seed
    P_cfx_w = np.copy(robot.P)
    P_cfx_w[:,0] = 0.0
    P_cfx_w[:,5:7] = 0.0    
    cfx_robot = rox.Robot(robot.H, P_cfx_w, robot.joint_type)

    tool = rr_current_tool
    payload = rr_current_payload
    payload_pose = rr_current_payload_pose

    setup_args = dict()
    if rr_mp.motion_setup_commands is not None:
        for setup_cmd in rr_mp.motion_setup_commands:
            #with suppress(OptionalCommandException):
                conv = get_command_conv(setup_cmd)
                conv.add_setup_args(setup_cmd, setup_args)
                if isinstance(conv,SetToolCommandConv):
                    tool = cmd_get_arg(setup_cmd,"tool_info")
                if isinstance(conv,SetPayloadCommandConv):
                    payload = cmd_get_arg(setup_cmd,"payload_info")
                    payload_pose = cmd_get_arg(setup_cmd,"payload_pose")
                
    if "tool" not in setup_args and rr_current_tool is not None:
        setup_args["tool"] = rr_tool_to_abb(rr_current_tool)
    if "gripload" not in setup_args and rr_current_payload is not None:
        setup_args["gripload"] = rr_payload_to_abb(rr_current_payload, rr_current_payload_pose)

    if rr_mp.extended is not None:
        first_cmd_num_rr = rr_mp.extended.get("first_command_number")
        setup_args["first_cmd_num"] = int(first_cmd_num_rr.data)
    mp = abb_exec.MotionProgram(**setup_args)
    for cmd in rr_mp.motion_program_commands:
        #with suppress(OptionalCommandException):
            apply_rr_motion_command_to_mp(cmd, mp, robot=robot, cfx_robot=cfx_robot)
        
    return mp, tool, payload, payload_pose